/*
 * Ali M560x driver
 *
 * Copyright (C) 2005 m560x Linux Driver Project <m560x@x3ng.com.br>
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License as
 *	published by the Free Software Foundation, version 2.
 *
 * This driver is based on the 2.6.3 version of drivers/usb/usb-skeleton.c
 * but has been rewritten to be easy to read and use, as no locks are now
 * needed anymore.
 *
 */

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/vmalloc.h>
#include <linux/kref.h>
#include <asm/uaccess.h>
#include <linux/videodev.h>
#include <linux/usb.h>                                                                                                     

#include "m560x.h"

/* module parameter */
static unsigned int debug = 0;

/* table of usb-devices that work with this driver */
static struct usb_device_id m560x_table [] = {
	{ USB_DEVICE(USB_M560X_VENDOR_ID, USB_M560X_PRODUCT_M5602) },
	{ USB_DEVICE(USB_M560X_VENDOR_ID, USB_M560X_PRODUCT_M5603c) },
	{ }					/* Terminating entry */
};
MODULE_DEVICE_TABLE (usb, m560x_table);
                              
static void m560x_delete(struct kref *kref)
{
	struct m560x_camera *cam = to_m560x_camera(kref);

	if (cam) {
		usb_put_dev(cam->udev);
		if (cam->int_in_buffer)
			kfree (cam->int_in_buffer);
		if (cam->isoc_in_buffer)
			kfree (cam->isoc_in_buffer);
		cam->udev = NULL;
			
		kfree (cam);
	}
}

static void *rvmalloc(unsigned long size)
{
	void *mem;
	unsigned long adr;

	size = PAGE_ALIGN(size);
	mem = vmalloc_32(size);
	if (!mem)
		return NULL;

	memset(mem, 0, size); /* Clear the ram out, no junk to the user */
	adr = (unsigned long) mem;
	while (size > 0) {
		SetPageReserved(vmalloc_to_page((void *)adr));
		adr += PAGE_SIZE;
		size -= PAGE_SIZE;
	}

	return mem;
}

static void rvfree(void *mem, unsigned long size)
{
	unsigned long adr;

	if (!mem)
		return;

	adr = (unsigned long) mem;
	while ((long) size > 0) {
		ClearPageReserved(vmalloc_to_page((void *)adr));
		adr += PAGE_SIZE;
		size -= PAGE_SIZE;
	}
	vfree(mem);
}

static int initialize_camera(struct m560x_camera *cam)
{
	PDEBUG(DBG_TRACE, "initialize_camera");

	info("initialize_camera - not implemented yet...");

	return 0;
}

static int set_camera_power(struct m560x_camera *cam, int state)
{
	PDEBUG(DBG_TRACE, "set_camera_power");

	info("set_camera_power - not implemented yet...");

	return 0;
}

/*
 *  ---------------------------------------------------------------------------
 * Video4Linux Functions
 * ---------------------------------------------------------------------------
 */
static int v4l_m560x_open(struct inode *inode, struct file *file)
{
	struct video_device *vdev = video_devdata(file);
	struct m560x_camera *cam =
	    (struct m560x_camera *) vdev->priv;
	int retval = 0;

	PDEBUG(DBG_TRACE, "v4l_m560x_open");

	if (!cam) {
		err("m560x video_device improperly initialized");
		retval = -ENODEV;
		goto error;
	}

	/* increment our usage count for the device */
	kref_get(&cam->kref);

	/* save our object in the file's private structure */
	file->private_data = cam;

	// First upload firmware, then turn the camera on
	if (!cam->is_initialized) {
		initialize_camera(cam);

		cam->is_initialized = 1;
	}

	set_camera_power(cam, 1);

	return 0;
	
error:
	if (cam)
		kref_put(&cam->kref, m560x_delete);

	return retval;
}

static int v4l_m560x_close(struct inode *inode, struct file *file)
{
	struct m560x_camera *cam = (struct m560x_camera *)file->private_data;
	int retval = 0;

	PDEBUG(DBG_TRACE, "v4l_m560x_close");

	if (!cam) {
		err("m560x video_device improperly initialized");
		retval = -ENODEV;
		goto error;
	}

	/* it's not the end of the world if
	 * we fail to turn the camera off.
	 */
	set_camera_power(cam, 0);

	/* decrement the count on our device */
	kref_put(&cam->kref, m560x_delete);

	return 0;

error:
	return retval;
}

static ssize_t v4l_m560x_read(struct file *file, char *buffer, size_t count, loff_t *ppos)
{
	PDEBUG(DBG_TRACE, "v4l_m560x_read");
	info("v4l_m560x_read - not implemented yet...");
	return 0;
}


static int v4l_m560x_mmap(struct file *file, struct vm_area_struct *vma)
{
	PDEBUG(DBG_TRACE, "v4l_m560x_mmap");
	info("v4l_m560x_mmap - not implemented yet...");
	return 0;
}

static int v4l_m560x_ioctl(struct inode *inode, struct file *file, unsigned int ioctlnr, unsigned long arg)
{
	void __user *user_arg = (void __user *)arg;
	int retval = 0;
	struct m560x_camera *cam = (struct m560x_camera *)file->private_data;

	PDEBUG(DBG_TRACE, "v4l_m560x_ioctl");
	
	if (!cam) {
		err("m560x video_device improperly initialized");
		retval = -ENODEV;
		goto exit;
	}

	switch (ioctlnr) {
		/* query capabilities */
	case VIDIOCGCAP:
		{
			struct video_capability *b = user_arg;

			info("v4l_m560x_ioctl - VIDIOCGCAP - not implemented yet");

			strcpy (b->name, cam->name);
			b->type = VID_TYPE_CAPTURE;
			b->channels = 1;
			b->audios = 0;
			b->maxwidth = cam->maxwidth;
			b->maxheight = cam->maxheight;
			b->minwidth = cam->maxwidth / 2;
			b->minheight = cam->maxheight / 2;
			break;
		}
		/* get/set video source - we are a camera and nothing else */
	case VIDIOCGCHAN:
		{
			info("v4l_m560x_ioctl - VIDIOCGCHAN - not implemented yet");
			retval = -EINVAL;
			break;
		}
	case VIDIOCSCHAN:
		{
			info("v4l_m560x_ioctl - VIDIOCSCHAN - not implemented yet");
			retval = -EINVAL;
			break;
		}

		/* image properties */
	case VIDIOCGPICT:
		{
			info("v4l_m560x_ioctl - VIDIOCGPICT - not implemented yet");
			retval = -EINVAL;
			break;
		}

	case VIDIOCSPICT:
		{
			info("v4l_m560x_ioctl - VIDIOCSPICT - not implemented yet");
			retval = -EINVAL;
			break;
		}

		/* get/set capture window */
	case VIDIOCGWIN:
		{
			info("v4l_m560x_ioctl - VIDIOCGWIN - not implemented yet");
			retval = -EINVAL;
			break;
		}

	case VIDIOCSWIN:
		{
			info("v4l_m560x_ioctl - VIDIOCSWIN - not implemented yet");
			retval = -EINVAL;
			break;
		}

		/* mmap interface */
	case VIDIOCGMBUF:
		{
			info("v4l_m560x_ioctl - VIDIOCGMBUF - not implemented yet");
			retval = -EINVAL;
			break;
		}

	case VIDIOCMCAPTURE:
		{
			info("v4l_m560x_ioctl - VIDIOCMCAPTURE - not implemented yet");
			retval = -EINVAL;
			break;
		}

	case VIDIOCSYNC:
		{
			info("v4l_m560x_ioctl - VIDIOCSYNC - not implemented yet");
			retval = -EINVAL;
			break;
		}

		/* pointless to implement overlay with this camera */
	case VIDIOCCAPTURE:
	case VIDIOCGFBUF:
	case VIDIOCSFBUF:
	case VIDIOCKEY:
		retval = -EINVAL;
		break;

		/* tuner interface - we have none */
	case VIDIOCGTUNER:
	case VIDIOCSTUNER:
	case VIDIOCGFREQ:
	case VIDIOCSFREQ:
		retval = -EINVAL;
		break;

		/* audio interface - we have none */
	case VIDIOCGAUDIO:
	case VIDIOCSAUDIO:
		retval = -EINVAL;
		break;
	default:
		retval = -ENOIOCTLCMD;
		break;
	}
	
exit:
	return retval;
}


static struct file_operations v4l_m560x_fops = {
	.owner		= THIS_MODULE,
	.open		= v4l_m560x_open,
	.release	= v4l_m560x_close,
	.read		= v4l_m560x_read,
	.mmap		= v4l_m560x_mmap,
	.ioctl		= v4l_m560x_ioctl,
	.llseek		= no_llseek,
};

static struct video_device v4l_m560x_template = {
	.owner 		= THIS_MODULE,
	.name 		= "Ali M560x-based USB Camera",
	.type 		= VID_TYPE_CAPTURE,
	.hardware 	= VID_HARDWARE_M560X,
	.fops 		= &v4l_m560x_fops,
	.release	= video_device_release,
	.minor 		= -1,
};

/****************************************************************************
 *  sysfs
 ***************************************************************************/
#define m560x_file(name, variable, field)                              \
static ssize_t show_##name(struct class_device *class_dev, char *buf)   \
{                                                                       \
        struct video_device *vdev = to_video_device(class_dev);         \
        struct m560x_camera *cam = video_get_drvdata(vdev);                  \
        return sprintf(buf, field, cam->variable);                      \
}                                                                       \
static CLASS_DEVICE_ATTR(name, S_IRUGO, show_##name, NULL);

m560x_file(model, name, "%s\n");
m560x_file(in_use, user, "%d\n");
m560x_file(streaming, streaming, "%d\n");
m560x_file(palette, palette, "%i\n");
m560x_file(frames_total, readcount, "%d\n");
m560x_file(frames_read, framecount, "%d\n");
m560x_file(packets_dropped, dropped, "%d\n");
m560x_file(decoding_errors, error, "%d\n");

static void m560x_create_sysfs_files(struct video_device *vdev)
{
        video_device_create_file(vdev, &class_device_attr_model);
        video_device_create_file(vdev, &class_device_attr_in_use);
        video_device_create_file(vdev, &class_device_attr_streaming);
        video_device_create_file(vdev, &class_device_attr_palette);
        video_device_create_file(vdev, &class_device_attr_frames_total);
        video_device_create_file(vdev, &class_device_attr_frames_read);
        video_device_create_file(vdev, &class_device_attr_packets_dropped);
        video_device_create_file(vdev, &class_device_attr_decoding_errors);
}

static void m560x_remove_sysfs_files(struct video_device *vdev)
{
        video_device_remove_file(vdev, &class_device_attr_model);
        video_device_remove_file(vdev, &class_device_attr_in_use);
        video_device_remove_file(vdev, &class_device_attr_streaming);
        video_device_remove_file(vdev, &class_device_attr_palette);
        video_device_remove_file(vdev, &class_device_attr_frames_total);
        video_device_remove_file(vdev, &class_device_attr_frames_read);
        video_device_remove_file(vdev, &class_device_attr_packets_dropped);
        video_device_remove_file(vdev, &class_device_attr_decoding_errors);
}

/****************************************************************************
 * USB - Functions
 ***************************************************************************/
static int usb_m560x_probe(struct usb_interface *interface, const struct usb_device_id *id)
{
	struct m560x_camera *cam = NULL;
	struct usb_device *dev = interface_to_usbdev(interface);
	struct usb_host_interface *iface_desc;
	struct usb_endpoint_descriptor *endpoint;
	char *camera_name = NULL;
	size_t buffer_size;
	int i;
	int retval = -ENOMEM;

	info("m560x_probe - M560x-based WebCam connected");

	/* allocate memory for our device state and initialize it */
	cam = kmalloc(sizeof(*cam), GFP_KERNEL);
	if (cam == NULL) {
		err("Out of memory");
		goto error;
	}
	memset(cam, 0x00, sizeof(*cam));
	kref_init(&cam->kref);

	cam->interface = interface;
	cam->udev = dev;

	if ((le16_to_cpu(dev->descriptor.idVendor) == USB_M560X_VENDOR_ID)
		&& (le16_to_cpu(dev->descriptor.idProduct) == USB_M560X_PRODUCT_M5602)) {
		camera_name = "Ali M5602-based WebCam";
		info("Ali M5602-based WebCam found.");
	}
	else if ((le16_to_cpu(dev->descriptor.idVendor) == USB_M560X_VENDOR_ID)
		&& (le16_to_cpu(dev->descriptor.idProduct) == USB_M560X_PRODUCT_M5603c)) {
		camera_name = "Ali M5603c-based WebCam";
		info("Ali M5603c-based WebCam found.");
	} else {
		err("Vendor/Product ID do not match known M506x WebCam devices.");
		retval = -ENODEV;
		goto error;
	}
	cam->name = camera_name;  

	/* set up the endpoint information */
	/* use only the first bulk-in and bulk-out endpoints */
	iface_desc = interface->cur_altsetting;
	for (i = 0; i < iface_desc->desc.bNumEndpoints; ++i) {
		endpoint = &iface_desc->endpoint[i].desc;

		if (!cam->int_in_endpointAddr
			&& ((endpoint->bEndpointAddress & USB_ENDPOINT_DIR_MASK) == USB_DIR_IN)
			&& ((endpoint->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK) == USB_ENDPOINT_XFER_INT)) {
			/* we found an int in endpoint */
			buffer_size = le16_to_cpu(endpoint->wMaxPacketSize);
			cam->int_in_size = buffer_size;
			cam->int_in_endpointAddr = endpoint->bEndpointAddress;
			cam->int_in_buffer = kmalloc(buffer_size, GFP_KERNEL);
			if (!cam->int_in_buffer) {
				err("Could not allocate int_in_buffer");
				goto error;
			}
		}

		if (!cam->isoc_in_endpointAddr
			&& ((endpoint->bEndpointAddress & USB_ENDPOINT_DIR_MASK) == USB_DIR_IN)
			&& ((endpoint->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK) == USB_ENDPOINT_XFER_ISOC)) {
			/* we found an isoc in endpoint */
			buffer_size = le16_to_cpu(endpoint->wMaxPacketSize);
			cam->isoc_in_size = buffer_size;
			cam->isoc_in_endpointAddr = endpoint->bEndpointAddress;
			cam->isoc_in_buffer = kmalloc(buffer_size, GFP_KERNEL);
			if (!cam->isoc_in_buffer) {
				err("Could not allocate isoc_in_buffer");
				goto error;
			}
		}
	}
	if (!(cam->int_in_endpointAddr && cam->isoc_in_endpointAddr)) {
		err("Could not find both int-in and isoc-in endpoints");
		goto error;
	}

	/* save our data pointer in this interface device */
	usb_set_intfdata(interface, cam);

	cam->vdev = video_device_alloc();
	if (!cam->vdev) {
		retval = -ENOMEM;
		goto error;
	}

	memcpy(cam->vdev, &v4l_m560x_template, sizeof (v4l_m560x_template));
	cam->vdev->priv = cam;	// sort of a reverse mapping for those functions that get vdev only
	cam->vdev->dev = &interface->dev;
	video_set_drvdata(cam->vdev, cam);
	
	memcpy (cam->vdev->name, cam->name, strlen (cam->name));
	init_waitqueue_head (&cam->wq);
	init_MUTEX (&cam->lock);
	wmb ();

	if (video_register_device(cam->vdev, VFL_TYPE_GRABBER, -1) == -1) {
		err("video_register_device failed");
		usb_set_intfdata(interface, NULL);
		video_device_release(cam->vdev);
		retval = -EIO;
		goto error;
	}

	info("m560x_probe - Ali_M560x WebCam driver is now controlling video device %d",cam->vdev->minor);

	m560x_create_sysfs_files(cam->vdev);

	return 0;

error:
	if (cam)
		kref_put(&cam->kref, m560x_delete);
	return retval;
}

static void usb_m560x_disconnect(struct usb_interface *interface)
{
	struct m560x_camera *cam = usb_get_intfdata(interface);
	int minor = cam->vdev->minor;

	/* prevent m560x_open() from racing m560x_disconnect() */
	lock_kernel();

	usb_set_intfdata(interface, NULL);
	if (cam) {
		/* We don't want people trying to open up the device */
		if (cam->vdev) {
			m560x_remove_sysfs_files(cam->vdev);
			video_unregister_device(cam->vdev);
			cam->vdev = NULL;
		}
	}
	
	info("m560x_disconnect - Ali_M560x WebCam driver no longer controlling video device %d", minor);

	unlock_kernel();

	/* decrement our usage count */
	kref_put(&cam->kref, m560x_delete);

	info("m560x_disconnect- M560x-based WebCam disconnected");;
}

static struct usb_driver usb_m560x_driver = {
//	.owner = THIS_MODULE,
	.name =	"m560x",
	.probe = usb_m560x_probe,
	.disconnect = usb_m560x_disconnect,
	.id_table = m560x_table,
};

static int __init usb_m560x_init(void)
{
	int result;

	info("usb_m560x_init - M560x-based WebCam driver startup");

	/* register this driver with the USB subsystem */
	result = usb_register(&usb_m560x_driver);
	if (result)
		err("usb_register failed. Error number %d", result);

	info(DRIVER_VERSION ":" DRIVER_DESC);
	
	return result;
}

static void __exit usb_m560x_exit(void)
{
	info("usb_m560x_exit - M560x-based WebCam driver shutdown");
	
	/* deregister this driver with the USB subsystem */
	usb_deregister(&usb_m560x_driver);
}


module_init (usb_m560x_init);
module_exit (usb_m560x_exit);

MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_LICENSE("GPL");
module_param(debug, int, S_IRUGO | S_IWUSR);
MODULE_PARM_DESC (debug, "Debug enabled or not");
